import ansiRegex from 'ansi-regex'
import superSplit from 'super-split'
import arrayUniq from 'array-uniq'
import stripAnsi from 'strip-ansi'

const realignOutOfBoundsCoords = (text, options) => {
	const plain = stripAnsi(text)
	const lines = plain.split('\n')
	const totalLines = lines.length

	// 'End.line marker out of bounds (max).'
	if (options.end.line > totalLines) {
		options.end.line = totalLines
	}

	// 'Start.line marker out of bounds (min).'
	if (options.start.line < 1) {
		options.start.line = 1
	}

	// 'Start.column marker out of bounds (min).'
	if (options.start.column < 1) {
		options.start.column = 1
	}

	// 'End.column marker out of bounds (max).'
	if (options.end.column > lines[options.end.line - 1].length) {
		options.end.column = lines[options.end.line - 1].length
	}

	if (options.start.line > options.end.line) {
		throw new Error('Your start line is after your end line.')
	}

	if (options.start.line === options.end.line
		&& options.end.column < options.start.column) {
		throw new Error('Your end column is after your start column.')
	}

	return false
}

// Returns arys:
// 1 - ANSI Escape sequences from section
// 2 - Glyphs in section (ansi escape seq - or - ascii character)
const atomize = section => {
	const ansies = arrayUniq(section.match(ansiRegex()))
	const words = superSplit(section, ansies)

	let glyphs = []
	if (Array.isArray(words)) {
		for (const word of words) {
			if (ansies.includes(word) === false) {
				glyphs = glyphs.concat(word.split(''))
				continue
			}

			glyphs.push(word)
		}
	}

	return {ansies, glyphs}
}

const markSection = (section, options, linear) => {
	const {ansies, glyphs} = atomize(section)

	let x = 0
	let y = 0
	let inPoint
	let outPoint
	let output = ''
	const height = options.end.line - options.start.line

	const markNotBegun = () => typeof inPoint !== 'number'
			&& typeof outPoint !== 'number'

	const markHasEnded = () => typeof inPoint === 'number'
			&& typeof outPoint === 'number'

	const outsideOfMark = () => markNotBegun() || markHasEnded()

	for (const glyph of glyphs) {
		if (ansies.includes(glyph) === false) {
			if (glyph === '\n' && !linear) {
				y += 1
				x = -1
			}

			x += 1

			if (x === options.start.column && y === 0) {
				inPoint = output.length
			}

			output += glyph

			if (x === options.end.column && y === height) {
				outPoint = output.length
			}

			continue
		}

		if (outsideOfMark()) {
			output += glyph
		} else if (!outsideOfMark() && !options.resetColor) {
			output += glyph
		}
	}

	const pre = output.slice(0, Math.max(0, inPoint))
	// eslint-disable-next-line unicorn/prefer-string-slice
	const mark = options.color(output.substr(inPoint, outPoint - inPoint))
	const post = output.slice(outPoint)
	const sectionMarked = pre + mark + post

	return sectionMarked
}

const mark2d = (text, options) => {
	realignOutOfBoundsCoords(text, options)

	const lines = text.split('\n')

	// Minus 1: because line and column numbers start at 1
	const startLine = options.start.line - 1
	const endLine = options.end.line - 1

	// Plus 1: because slice does not include the end indice
	const unmarkedSection = lines.slice(startLine, endLine + 1).join('\n')

	const preSection = lines.slice(0, startLine)
	const markedSection = markSection(unmarkedSection, options)
	const postSection = lines.slice(endLine + 1)

	const result = preSection.concat([markedSection]).concat(postSection).join('\n')
	return result
}

const mark1d = (text, options, linear) => {
	const markedSection = markSection(text, options, linear)
	return markedSection
}

const mark = (text, options, linear) => linear ? mark1d(text, options, linear) : mark2d(text, options)

const validMarkersNumbers = options => typeof options.start === 'number'
		&& typeof options.end === 'number'

const validMarkersObject = options => typeof options.start === 'object'
		&& typeof options.end === 'object'
		&& typeof options.start.line === 'number'
		&& typeof options.start.column === 'number'
		&& typeof options.end.line === 'number'
		&& typeof options.end.column === 'number'

const ansiMark = (text, options) => {
	if (validMarkersObject(options)) {
		return mark(text, options)
	}

	if (validMarkersNumbers(options)) {
		options.start = {line: 1, column: options.start}
		options.end = {line: 1, column: options.end}
		const linear = true
		return mark(text, options, linear)
	}

	throw new Error('Invalid marker definition.')
}

export default ansiMark
